@@ -9,6 +9,8 @@ Puppet::Type.type(:keystone_endpoint).provide(
 
   include PuppetX::Keystone::CompositeNamevar::Helpers
 
+  attr_accessor :property_hash, :property_flush
+
   @endpoints     = nil
   @services      = nil
   @credentials   = Puppet::Provider::Openstack::CredentialsV3.new
@@ -31,30 +33,17 @@ Puppet::Type.type(:keystone_endpoint).provide(
     if self.class.do_not_manage
       fail("Not managing Keystone_endpoint[#{@resource[:name]}] due to earlier Keystone API failures.")
     end
-    # Reset the cache.
-    self.class.services = nil
     name   = resource[:name]
     region = resource[:region]
     type   = resource[:type]
-    type   = self.class.type_from_service(name) unless set?(:type)
     @property_hash[:type] = type
-    services = self.class.services.find_all { |s| s[:name] == name }
-    service = services.find { |s| s[:type] == type }
-
-    if service.nil? && services.count == 1
-      # For backward comptatibility, match the service by name only.
-      name = services[0][:id]
-    else
-      # Math the service by id.
-      name = service[:id] if service
-    end
     ids = []
-
+    s_id = service_id
     created = false
     [:admin_url, :internal_url, :public_url].each do |scope|
       if resource[scope]
         created = true
-        ids << endpoint_create(name, region,  scope.to_s.sub(/_url$/, ''),
+        ids << endpoint_create(s_id, region, scope.to_s.sub(/_url$/, ''),
                                resource[scope])[:id]
       end
     end
@@ -75,7 +64,7 @@ Puppet::Type.type(:keystone_endpoint).provide(
     end
     ids = @property_hash[:id].split(',')
     ids.each do |id|
-      self.class.request('endpoint', 'delete', id)
+      self.class.system_request('endpoint', 'delete', id)
     end
     @property_hash.clear
   end
@@ -134,11 +123,11 @@ Puppet::Type.type(:keystone_endpoint).provide(
       new(
         :name         => endpoint[:name],
         :ensure       => :present,
-        :id           => "#{endpoint[:admin][:id]},#{endpoint[:internal][:id]},#{endpoint[:public][:id]}",
-        :region       => endpoint[:admin][:region],
-        :admin_url    => endpoint[:admin][:url],
-        :internal_url => endpoint[:internal][:url],
-        :public_url   => endpoint[:public][:url]
+        :id           => make_id(endpoint),
+        :region       => get_region(endpoint),
+        :admin_url    => get_url(endpoint, :admin),
+        :internal_url => get_url(endpoint, :internal),
+        :public_url   => get_url(endpoint, :public)
       )
     end
   end
@@ -153,26 +142,33 @@ Puppet::Type.type(:keystone_endpoint).provide(
   end
 
   def flush
-    if @property_flush && @property_hash[:id]
-      ids = @property_hash[:id].split(',')
-      if @property_flush[:admin_url]
-        self.class.request('endpoint', 'set', [ids[0], "--url=#{resource[:admin_url]}"])
-      end
-      if @property_flush[:internal_url]
-        self.class.request('endpoint', 'set', [ids[1], "--url=#{resource[:internal_url]}"])
-      end
-      if @property_flush[:public_url]
-        self.class.request('endpoint', 'set', [ids[2], "--url=#{resource[:public_url]}"])
+    if property_flush && property_hash[:id]
+      scopes = [:admin_url, :internal_url, :public_url]
+      ids = Hash[scopes.zip(property_hash[:id].split(','))]
+      scopes.each do |scope|
+        if property_flush[scope]
+          if ids[scope].nil? || ids[scope].empty?
+            ids[scope] = endpoint_create(service_id, resource[:region],
+                                         scope.to_s.sub(/_url$/, ''),
+                                         property_flush[scope])[:id]
+          else
+            self.class.system_request('endpoint',
+                                      'set',
+                                      [ids[scope], "--url=#{resource[scope]}"])
+          end
+        end
       end
+      @property_hash = resource.to_hash
+      @property_hash[:id] = scopes.map { |s| ids[s] }.join(',')
+      @property_hash[:ensure] = :present
     end
-    @property_hash = resource.to_hash
   end
 
   private
 
   def endpoint_create(name, region, interface, url)
     properties = [name, interface, url, '--region', region]
-    self.class.request('endpoint', 'create', properties)
+    self.class.system_request('endpoint', 'create', properties)
   end
 
   private
@@ -181,7 +177,7 @@ Puppet::Type.type(:keystone_endpoint).provide(
     return @endpoints unless @endpoints.nil?
     prev_do_not_manage = self.do_not_manage
     self.do_not_manage = true
-    @endpoints = request('endpoint', 'list')
+    @endpoints = system_request('endpoint', 'list')
     self.do_not_manage = prev_do_not_manage
     @endpoints
   end
@@ -194,7 +190,7 @@ Puppet::Type.type(:keystone_endpoint).provide(
     return @services unless @services.nil?
     prev_do_not_manage = self.do_not_manage
     self.do_not_manage = true
-    @services = request('service', 'list')
+    @services = system_request('service', 'list')
     self.do_not_manage = prev_do_not_manage
     @services
   end
@@ -203,53 +199,51 @@ Puppet::Type.type(:keystone_endpoint).provide(
     @services = value
   end
 
-  def self.endpoint_from_region_name(region, name)
-    endpoints.find_all { |e| e[:region] == region && e[:service_name] == name }
-      .map { |e| e[:service_type] }.uniq
-  end
-
-  def self.type_from_service(name)
-    types = services.find_all { |s| s[:name] == name }.map { |e| e[:type] }.uniq
-    if types.count == 1
-      types[0]
-    else
-      # We don't fail here as it can happen during a ensure => absent.
-      PuppetX::Keystone::CompositeNamevar::Unset
-    end
-  end
-
-  def self.service_type(services, region, name)
-    nbr_of_services = services.count
-    err_msg         = ["endpoint matching #{region}/#{name}:"]
-    type            = nil
-
-    case
-    when nbr_of_services == 1
-      type = services[0]
-    when nbr_of_services > 1
-      err_msg += [endpoint_from_region_name(region, name).join(' ')]
-    when nbr_of_services < 1
-      # Then we try to get the type by service name.
-      type = type_from_service(name)
-    end
-
-    if !type.nil?
-      type
-    else
-      fail(Puppet::Error, 'Cannot get the correct endpoint type: ' \
-           "#{err_msg.join(' ')}")
-    end
-  end
-
   def self.transform_name(region, name, type)
-    if type == PuppetX::Keystone::CompositeNamevar::Unset
-      type = service_type(endpoint_from_region_name(region, name), region, name)
+    "#{region}/#{name}::#{type}"
+  end
+
+  def self.make_id(endpoint)
+    id_str = ''
+    id_sep = ''
+    [:admin, :internal, :public].each do |type|
+      id_str += id_sep
+      id_str += endpoint[type][:id] if endpoint[type]
+      id_sep = ','
     end
-    if type == PuppetX::Keystone::CompositeNamevar::Unset
-      Puppet.debug("Could not find the type for endpoint #{region}/#{name}")
-      "#{region}/#{name}"
+    id_str
+  end
+
+  def self.get_region(endpoint)
+    type = [:admin, :internal, :public].detect { |t| endpoint.key? t }
+    type ? endpoint[type][:region] : ''
+  end
+
+  def self.get_url(endpoint, type, default='')
+    endpoint[type][:url] rescue default
+  end
+
+  def service_id
+    # Reset the cache.
+    self.class.services = nil
+    name = resource[:name]
+    type = resource[:type]
+
+    services = self.class.services.find_all { |s| s[:name] == name }
+    service = services.find { |s| s[:type] == type }
+    service_id = ''
+    if service.nil? && services.count == 1
+      # For backward compatibility, match the service by name only.
+      service_id = services[0][:id]
     else
-      "#{region}/#{name}::#{type}"
+      # Math the service by id.
+      service_id = service[:id] if service
     end
+    if service_id.nil? || service_id.empty?
+      title = self.class.transform_name(resource[:region], resource[:name], resource[:type])
+      fail(Puppet::Error, "Cannot find service associated with #{title}")
+    end
+
+    service_id
   end
 end
